arXiv is now an independent nonprofit! Learn more
License: CC BY 4.0
arXiv:2502.09720v3 [cs.LG] 26 Jul 2025

NestQuant: Nested Lattice Quantization for Matrix Products and LLMs

Semyon Savkin11 1 Equal contribution Affiliation: MIT Email: semyon@mit.edu    Eitan Porat11 1 Equal contribution Affiliation: Independent Email: ethan.porat@gmail.com    Or Ordentlich Affiliation: Hebrew University of Jerusalem Email: or.ordentlich@mail.huji.ac.il    Yury Polyanskiy Affiliation: MIT Email: yp@mit.edu
Abstract

Post-training quantization (PTQ) has emerged as a critical technique for efficient deployment of large language models (LLMs). This work proposes NestQuant, a novel PTQ scheme for weights and activations that is based on self-similar nested lattices. Recent works have mathematically shown such quantizers to be information-theoretically optimal for low-precision matrix multiplication. We implement a practical low-complexity version of NestQuant based on Gosset lattice, making it a drop-in quantizer for any matrix multiplication step (e.g., in self-attention, MLP etc). For example, NestQuant quantizes weights, KV-cache, and activations of Llama-3-8B to 4 bits, achieving perplexity of 6.6 on wikitext2. This represents more than 55% reduction in perplexity gap with respect to unquantized model (perplexity of 6.14) compared to state-of-the-art Meta’s SpinQuant (perplexity 7.3), OstQuant (7.3) and QuaRot (8.2). Comparisons on bigger models (up to 70B) and on various LLM evaluation benchmarks confirm uniform superiority of NestQuant.

   

1 Introduction

Figure 1: Perplexity vs number of bits/entry in quantized Llama-3-8B for three regimes (weight-only, weights + KV cache, end-to-end) on wikitext2 at 2048 context.

There are three principal goals of post-training quantization (PTQ). First, reducing the number of bits per parameter allows for loading big models on cheap GPUs with limited memory, thus democratizing access to LLMs. This requires “weights-only” quantization algorithms of which the most popular are AWQ, GPTQ, and QuIP (see references in Section 2.2).

The second goal of PTQ is to accelerate inference. In LLMs most of the compute is spent multiplying matrices. Multiplying a pair of n×nn\times n matrices requires 2​n32n^{3} FLOPs and 38​R​n2{\frac{3}{8}}Rn^{2} bytes to exchange between the core and memory (here and below RR designates the number of bits required to store each entry of a vector/matrix). So when matrices are large (such as during the pre-fill phase when the prompt is processed) the GPU is compute-bound, while when nn is small (such as during generation) the GPU becomes memory-bound. One therefore needs to reduce RR by quantizing both weights and the KV cache, thereby alleviating the memory bandwidth bottleneck during generation.

The third goal of PTQ is to accelerate inference of giant LLMs that require hosting each layer on a separate GPU (pipelining parallelism). For this goal one needs to quantize activations passed from one layer to the next to reduce the communication bottleneck.

While quantization of weights to R=3,4R=3,4 and even R=2R=2 bits has been achieved with minimal loss of quality, quantization of KV cache and activations has been much more challenging. Popular algorithms for full quantization are LLM.int8(), SmoothQuant and SpinQuant (see references in Section 2.2), the latter having state-of-the-art performance. This work proposes an alternative algorithm (NestQuant) for quantizing weights, KV-cache, and activations. The algorithm is motivated by recent theoretical work on approximate matrix multiplication and follows several classical ideas such as Conway and Sloane’s Voronoi codes and their algorithms for finding the closest lattice vector in some canonical lattices in low/moderate dimensions.

Throughout this paper, NestQuant is described with the 88-dimensional Gosset lattice, but the framework is general and can be similarly implemented using any other base lattice, in any dimension, provided that a low-complexity closest lattice vector algorithm exists for this lattice.

1.1 Summary of results

Model Bits ↓\downarrow Bits (no zstd) ↓\downarrow ARC-C ↑\uparrow ARC-E ↑\uparrow Hellaswag ↑\uparrow PIQA ↑\uparrow Winogrande ↑\uparrow Zero-shot Avg ↑\uparrow Wikitext2 ppl ↓\downarrow Baseline (FP16) 16 16 0.54 0.78 0.79 0.81 0.74 0.73 6.1 Weights only LLM-QAT 4.00 - 0.51 0.77 0.48 0.79 0.72 0.65 7.7 GPTQ 4.00 - 0.47 0.72 0.74 0.77 0.71 0.68 7.2 SpinQuant 4.00 - 0.54 0.77 0.78 0.80 0.72 0.72 6.5 NestQuant q=14,k=4q=14,k=4 (ours) 3.99 4.06 0.53 0.78 0.79 0.80 0.73 0.72 6.3 Weights + KV cache SpinQuant 4.00 - 0.51 0.77 0.77 0.78 0.69 0.70 6.6 NestQuant q=14,k=4q=14,k=4 (ours) 3.99 4.06 0.53 0.78 0.79 0.79 0.74 0.72 6.4 Weights, KV cache, activations LLM-QAT 4.00 - 0.27 0.41 0.38 0.60 0.53 0.44 52.5 Quarot 4.00 - 0.44 0.67 0.75 0.75 0.66 0.67 8.4 SpinQuant 4.00 - 0.51 0.75 0.75 0.77 0.66 0.68 7.3 NestQuant q=14,k=4q=14,k=4 (ours) 3.99 4.06 0.51 0.75 0.78 0.79 0.72 0.71 6.6

Table 1: 4-bit quantization of Llama-3-8B. The bits column for NestQuant corresponds to actually measured average number of bits per entry (when a vector of auxiliary scaling coefficients β\beta is compressed via zstd) and the second column shows quantization rate when no compression step is used.

The NestQuant algorithm is described in Section 4. NestQuant is a generic drop-in replacement for any matrix multiplication. Its performance for synthetic random Gaussian matrices comes pretty close to information-theoretic limits (see Fig. 3) and significantly outperforms uniform quantization employed by SpinQuant. Switching from a scalar (uniform) quantization to vector quantization requires some price to pay computationally (Section E.1), however, among vector quantizers NestQuant is rather economical as it builds upon the Voronoi Codes framework of (1). We mention that in the presence of activation quantization, it is important to quantize weights properly – an innovation we call QA-LDLQ (Section 4.5).

Applying NestQuant to quantizing an actual LLMs (Llama 2 and 3 with parameter count ranging from 1B to 70B) shows massive end-to-end improvement: Fig. 1 shows a significant reduction of perplexity compared to the previous methods; Table 1 confirm enhanced performance on standard LLM benchmarks; and Table 2 show NestQuant stays consistently on top for different model sizes. In fact, NestQuant in full quantization (weights, activations, and KV cache) outperforms SOTA results with quantized weights and activations (but not KV cache).

The main source of improvement of NestQuant is demonstrated in Fig. 2 (although NestQuant uses an 8-dimensional Gosset lattice, not a 2D hexagonal one). More details on this as well as directions for improvement are discussed in Section 3.

Thus, we believe that NestQuant offers an excellent alternative to other algorithms. It quantizes weights, KV-cache and activations, and achieves significant improvement on both synthetic and real data.

1.2 Paper organization

We start with a detailed review of classical work on vector quantization and modern LLM quantization (Section 2). Then in Section 3 we explain the motivation for each step of the algorithm. Section 4 contains the pseudocode of the algorithm and diagram of quantized LLM. Finally, Section 5 concludes with details about empirical performance. Further details and evaluations are relegated to the Appendices.

2 Prior work

We briefly survey prior work, which we separate into work by information theorists and by the ML community.

2.1 Information-theoretic quantization

Rate RR quantization of an information source XX in ℝn\mathbb{R}^{n} is the operation of encoding it to n​RnR bits, from which a decoder can produce a reconstruction X^∈ℝn\hat{X}\in\mathbb{R}^{n} that has small distortion with respect to XX. The most popular distortion criterion is the quadratic loss, where the expected distortion is defined as D=1n​𝔼​‖X−X^‖2D=\frac{1}{n}\mathbb{E}\|X-\hat{X}\|^{2}, and here we restrict attention to this loss. Characterization of the optimal tradeoff between RR and DD is a classic topic in information theory, e.g. (2, Part V).

For a Gaussian source X∼𝒩⁡(0,In)X\sim\mathcal{N}(0,I_{n}) the rate-distortion theorem states that any compressor at rate RR must satisfy D≥D⁡(R)≜2−2​RD\geq D(R)\triangleq 2^{-2R}. Furthermore, as dimension nn increases there exist quantizers with distortion approaching D⁡(R)D(R). Notably, such quantizers can be made universal, in the sense that they attain distortion D⁡(R)D(R) not only for iid Gaussian XX but for any (even adversarial) input as long as its Euclidean norm is (1+o⁡(1))​n(1+o(1))\sqrt{n}.

One way for constructing these universal quantizers is based on lattices [3] that admit much more structure than more classical random codes (and ϵ\epsilon-nets).

Arguably, the most notable lattice-based quantization scheme is the family of Voronoi codes [1], which we use in this work.

How does one convert a quantizer adapted to Gaussian inputs to work (with the same guaranteed loss) on non-Gaussian data? In a nutshell, the idea is simple: if UU is chosen to be a random n×nn\times n orthogonal matrix then the entries of U​XUX will be distributed as iid Gaussian [4]. This idea of applying random rotations to smooth out the distribution of the quantizer’s input may be viewed as a special case of high-dimensional companding [5], and has been been used in various applications, such as image compression [6], and as a potential replacement for dithering [7], to name a few.

In the context of LLMs, the goal in quantization is slightly different since quantization is used to facilitate approximate matrix multiplication with reduced burden on the memory bandwidth. For example, when quantizing two vectors X,Y∈ℝnX,Y\in\mathbb{R}^{n} the goal is not to approximate them but to approximate their inner product. Recently, information-theoretic characterization of this task was completed in (8). Specifically, the authors show that if X,Y∼𝒩⁡(0,In)X,Y\sim\mathcal{N}(0,I_{n}) are independent then for any algorithm operating on the basis of rate-RR quantized representations of XX and YY we must have

𝔼​(X⊤​Y−X⊤​Y^)2≥n​Γ​(R),\mathbb{E}(X^{\top}Y-\widehat{X^{\top}Y})^{2}\geq n\Gamma(R), (1)

where X⊤​Y^\widehat{X^{\top}Y} is the reconstructed inner product, and

Γ⁡(R)={1−(1−(2⋅2−2​R∗−2−4​R∗))​RR∗R<R∗2⋅2−2​R−2−4​RR≥R∗,\displaystyle\Gamma(R)=\begin{cases}1-\left(1-(2\cdot 2^{-2R^{*}}-2^{-4R^{*}})\right)\frac{R}{R^{*}}&R<R^{*}\\ 2\cdot 2^{-2R}-2^{-4R}&R\geq R^{*}\end{cases}, (2)

where R∗≈0.906R^{*}\approx 0.906 is a solution to a certain transcendental fixed-point equation.

The same paper also constructs universal quantizers based on nested lattices that asymptotically (as n→∞n\to\infty) achieve this lower bound. Note that extension from vectors to matrices can be made trivially by observing that one can quantize each column separately and treat matrix product as a collection of inner products.

Remark 2.1.

Using optimal quantizers, in terms of achieving D⁡(R)D(R), for quantizing XX and YY, and then computing the inner product of the quantized vectors (maybe with some MMSE scaling) will not necessarily lead to the optimal performance Γ⁡(R)\Gamma(R). The reason is that the covariance matrix of each vector’s quantization error affects the inner product distortion through its trace, and its Frobenius norm [8]. For optimal lattice quantizers in any dimension, the Frobenius norm is guaranteed to be minimal [9], and consequently, using lattice quantizers seems essential for achieving Γ⁡(R)\Gamma(R).

In this work we show that with appropriate tweaks Voronoi codes can indeed result in practical fast and efficient algorithms for LLM quantization. We emphasize that most of our work is on simply developing a drop-in replacement for quantized matrix product and as such is not specific to LLMs.

The idea of applying random rotations to ‘‘Gaussianize’’ inputs in the context of approximate inner product computation is even more natural than in the standard quantization. Indeed, since one is only interested in the inner product, not vectors themselves, one does not need to store (even a seed used to generate the) random orthogonal matrix. This has been long exploited in locality-sensitive hashing (LSH) algorithms,** * In LSH, one typically performs several random projections of the vector and quantizes them. This is equivalent to performing random rotation and quantizing only a small number of entries of the rotated vector. which can be viewed as an (extremely low rate) quantizers for inner product approximation [10, 11, 12]. Unsurprisingly, as we will see next, random rotations have also been found quite useful for quantizing LLMs.

2.2 LLM quantization

One of the directions of prior research on LLM quantization is addressing the issue of activation outliers that hinder the quantization quality. These outliers are present in certain dimensions of activations, weights and KV cache. In LLM.int8() of [13], these outlier dimension are kept unquantized. In SmoothQuant [14] authors balance the scale of outliers between weights and activations by modifying LayerNorm’s diagonal matrices.

Going to random rotations, by rewriting matrix product A​B=(A​U)​(U⊤​B)AB=(AU)(U^{\top}B) for an orthogonal matrix UU, one gets matrices with much more Gaussian entries (few outliers) and can apply standard quantization algorithms. Some of the multiplications by UU can be merged with the weights (i.e. do not require additional runtime FLOPs), while the rest are applied at runtime. For the latter, matrices UU should have structure to enable fast multiplication. For example, QuaRot [15] uses randomized Hadamard matrices as coordinate transformations, which can be applied to a vector of size nn in O⁡(n​log⁡n)O(n\log n) additions. SpinQuant [16] uses a rotation parametrization with four orthogonal matrices R1R_{1}, R2R_{2}, R3R_{3}, R4R_{4}, where R1R_{1} and R2R_{2} can be arbitrary orthonormal matrices, and R3R_{3} and R4R_{4} should have a fast multiplication algorithm. The authors use Cayley SGD [17] to optimize R1R_{1} and R2R_{2} for minimization of the quantization error, while the matrices R3R_{3} and R4R_{4} are chosen to be random Hadamard.

Starting from LLM.int8() most of the schemes used uniform quantization (i.e. where a floating point vector simply rounded to a nearest integer after an appropriate rescaling). To the best of our knowledge, so far non-scalar quantization has only been used for weight-only compression for LLMs in QuIP# [18], which uses E8P codebook for 2-bit quantization, and applies Resdidual Vector Quantization [19] to get a 4-bit quantization scheme; and QTIP [20] which uses trelis codebook. Unfortunately these methods appear to be too expensive to apply in runtime, perhaps explaining why non-uniform quantization for activations and KV-cache was not attempted before this work.

Finally, when quantizing weight matrices, one may notice that MSE distortion loss should be replaced by a weighted-MSE loss dependent on the statistics of the incoming activations. We refer to this type of algorithms as LDLQ, following authors of QuIP (21), QuIP# (18) and GPTQ (22). In the presence of activation quantization, however, the LDLQ needs to be modified (see Section 4.5 introducing QA-LDLQ). Equipping NestQuant with QA-LDLQ incurs significant improvement; see results in Section 5.

Refer to caption
Refer to caption
Figure 2: Demonstrating advantage of NestQuant in 2D. Typical weights and activations are vectors inside the black circle. Uniform quantization wastes about 32% of allocated bitstrings for vectors outside of the circle, while nested hexagonal lattices only wastes 15% (explicitly enumerating points inside the circle to avoid the waste is too slow to do at runtime). This allows NestQuant to use finer grid while quantizing to the same rate RR. The gain becomes much more dramatic in higher dimensions.

3 Outline of NestQuant approach

In this section we outline the main components of our approach. A detailed description is brought in the following section.

When designing a quantizer, one needs to make some assumptions about the distribution of the source that will be fed to it. While weights (and sometimes activations) can be well-approximated by Gaussians, their magnitudes are wildly varied. Thus, one employs two ideas: normalization and random rotation.

Normalization: In most of the literature, the normalization is done by taking an input vector of large dimension nn (e.g. n=4096n=4096 for Llama-3), dividing by the L∞L_{\infty} norm to get entries to be in [−1,1][-1,1] and then applying uniform quantization. This is suboptimal for two reasons: one is that uniform quantization induces error that is distributed uniformly on the small cube, which is suboptimal from the MSE point of view. Second reason, much more serious, is known as the shaping gain and demonstrated on Fig. 2. When entries of the vector are Gaussian, it will typically lie inside the black circle. Thus those grid elements outside of it will almost never be used, wasting bitspace.

Instead, we use normalization by the L2L_{2}-norm (see Algorithm 3) and then use points inside the Voronoi region of a Gosset lattice, which as Fig. 2 (right) demonstrates wastes a lot fewer bitstrings for rare vectors, thus allowing us to use finer grids.

Random rotation: When input to the quantizer significantly differs from the presumed model (of iid Gaussians), performance can become quite poor. As discussed previously, multiplying by a random orthoghonal matrix UU provides an excellent fix. Specifically, U​XUX vector becomes uniform on the nn-sphere of radius n\sqrt{n}, and small chunks of this vector have distribution very close to Gaussian iid. In particular, the total variation between any subset of dd coordinates and 𝒩⁡(0,Id)\mathcal{N}(0,I_{d}) is O⁡(d2/n)O(d^{2}/n) [4], such that for d=o⁡(n)d=o(\sqrt{n}) what we quantize is effectively iid Gaussian.

Lattices: A lattice Λ⊂ℝd\Lambda\subset\mathbb{R}^{d} is a discrete subgroup of ℝd\mathbb{R}^{d}. Any lattice Λ⊂ℝd\Lambda\subset\mathbb{R}^{d} has a (non-unique) generating matrix G∈ℝd×dG\in\mathbb{R}^{d\times d}, such that Λ=G​ℤd\Lambda=G\mathbb{Z}^{d} (that is, Λ\Lambda is the integer span of the columns of GG). We define the nearest neighbor quantizer QΛ:ℝd→ΛQ_{\Lambda}:\mathbb{R}^{d}\to\Lambda as

QΛ​(x)=arg⁡minλ∈Λ⁡‖x−λ‖,\displaystyle Q_{\Lambda}(x)=\argmin_{\lambda\in\Lambda}\|x-\lambda\|, (3)

where ties are broken arbitrarily, but in a systematic manner. The Voronoi region 𝒱Λ\mathcal{V}_{\Lambda} is defined as the set of all points in ℝd\mathbb{R}^{d} that are closer to 00 than to any other lattice point

𝒱Λ={x∈ℝd:QΛ​(x)=0}.\displaystyle\mathcal{V}_{\Lambda}=\left\{x\in\mathbb{R}^{d}~:~Q_{\Lambda}(x)=0\right\}. (4)

The covolume of a lattice is defined as covol⁡(Λ)=|detG|=vol⁡(𝒱Λ)\mathrm{covol}(\Lambda)=|\det G|=\mathrm{vol}(\mathcal{V}_{\Lambda}). We say that a lattice Λc⊂ℝd\Lambda_{c}\subset\mathbb{R}^{d} is nested in the lattice Λf⊂ℝd\Lambda_{f}\subset\mathbb{R}^{d}, if Λc⊂Λf\Lambda_{c}\subset\Lambda_{f}. Note that for any integer q≥2q\geq 2 we have that q​Λ⊂Λq\Lambda\subset\Lambda, and that Λ/q​Λ≅(ℤ/q​ℤ)d\Lambda/q\Lambda\cong(\mathbb{Z}/q\mathbb{Z})^{d}. For an introduction to lattices and their role in quantization, see [3].

Complexity of lattice quantization: In order to explain our choice of nested lattice quantizer, we need to carefully balance several requirements. One of the key ones is complexity. It is known that finding (even approximating) a nearest lattice point is a well-known cryptographic assumption [23]. Thus, we are not suggesting to directly operate on nn-dimensional lattices. Instead, we partition the nn-vector into sections, each of dimension dd and apply lattice quantization to dd-subvectors. Equivalently, our vector quantizers for ℝn\mathbb{R}^{n} are constructed as Cartesian products of vector quantizers of small dimension dd (we will take d=8d=8 for all experiments).

Granular and overload quantization errors: There are two different sources of errors for lattice quantizers. The first is called granular quantization error, and is related to the second moment of the lattice Voronoi region. A common way to measure the granular error corresponding to a lattice Λ⊂ℝd\Lambda\subset\mathbb{R}^{d} is via the normalized second moment (NSM) defined as

G⁡(Λ)=1vol​(𝒱Λ)1+2d​1d​∫x∈𝒱Λ‖x‖2​𝑑x.\displaystyle G(\Lambda)=\frac{1}{\mathrm{vol}(\mathcal{V}_{\Lambda})^{1+\frac{2}{d}}}\frac{1}{d}\int_{x\in\mathcal{V}_{\Lambda}}\|x\|^{2}dx. (5)

This quantity corresponds to the MSE when Λ\Lambda is normalized to have unit covolume and is then used as a quantizer. It is well known that the optimal (smallest) NSM among all lattices in ℝd\mathbb{R}^{d} approaches 12​π​e\frac{1}{2\pi e} from above as dd increases [3]. Furthermore, for d=1d=1 we get G⁡(ℤ)=112G(\mathbb{Z})=\frac{1}{12}. Consequently, in terms of granular error, by using high-dimensional lattices instead of the simple scalar quantizer based on ℤ\mathbb{Z} we can already gain a factor of 2​π​e12≈1.42329\frac{2\pi e}{12}\approx 1.42329 in performance (the Gosset lattice achieves 1.161.16 gain).

Notice, however, that representing xx as QΛ​(x)Q_{\Lambda}(x), the nearest lattice point in Λ\Lambda, requires infinitely many bits, since the lattice is infinite. Since we only have 2d​R2^{dR} bitstrings to allocate, we need to select a subset of Λ\Lambda that will be actually used. Selection of a body 𝒮⊂ℝd\mathcal{S}\subset\mathbb{R}^{d} so that |Λ∩𝒮|=2d​R|\Lambda\cap\mathcal{S}|=2^{dR} is called shaping. If QΛ​(x)∈𝒮Q_{\Lambda}(x)\in\mathcal{S}, then the quantization error x−QΛ​(x)x-Q_{\Lambda}(x) will be in 𝒱Λ\mathcal{V}_{\Lambda} and we will only suffer from a granular error. However, when QΛ​(x)∉𝒮Q_{\Lambda}(x)\notin\mathcal{S} the quantization error is no longer in 𝒱Λ\mathcal{V}_{\Lambda} and may be have much greater magnitude than a typical granular error. We refer to those type of errors, where QΛ​(x)∉𝒮Q_{\Lambda}(x)\notin\mathcal{S}, as overload errors.

Generally speaking, in order to achieve a small quantization error, one must keep the probability of overload very small. This can be attained by scaling up the codebook to β​𝒞=β​Λ∩β​𝒮\beta\mathcal{C}=\beta\Lambda\cap\beta\mathcal{S} with a large enough β>0\beta>0 such that overload becomes very rare. However, increasing β\beta also increases the squared granular error by a factor of β2\beta^{2}. Thus, one would like to use the smallest possible β\beta for which overload is rare. In order to allow for smaller β\beta, we would like to choose 𝒮⊂ℝn\mathcal{S}\subset\mathbb{R}^{n} such that β​𝒮\beta\mathcal{S} captures as much Gaussian mass as possible.

Denote by μ=𝒩⁡(0,Id)\mu=\mathcal{N}(0,I_{d}) the standard Gaussian measure. Since we need 2d​R=|Λ∩𝒮|≈vol⁡(𝒮)covol⁡(Λ)2^{dR}=|\Lambda\cap\mathcal{S}|\approx\frac{\mathrm{vol}(\mathcal{S})}{\mathrm{covol}(\Lambda)}, a good shaping region 𝒮\mathcal{S} maximizes μ⁡(𝒮)\mu(\mathcal{S}), which in turn minimizes the overload probability that is approximated by 1−μ⁡(𝒮)1-\mu(\mathcal{S}), under a volume constraint. Clearly, the optimal 𝒮\mathcal{S} under this criterion is r​ℬr\mathcal{B} where ℬ={x∈ℝd:‖x‖≤reff​(1)}\mathcal{B}=\{x\in\mathbb{R}^{d}~:~\|x\|\leq r_{\text{eff}}(1)\} is a Euclidean ball with radius reff​(1)r_{\text{eff}}(1) chosen such that vol⁡(ℬ)=1\mathrm{vol}(\mathcal{B})=1, and rr is chosen such that vol⁡(r​ℬ)=rd\mathrm{vol}(r\mathcal{B})=r^{d} satisfies the required volume constraint. Unfortunately, for d>1d>1 the codebook 𝒞=Λ∩r​ℬ\mathcal{C}=\Lambda\cap r\mathcal{B} loses much of the lattice structure, and does not admit an efficient enumeration, and consequently encoding and decoding require using a lookup table (LUT). QuIP# used this approach with Λ=E8\Lambda=E_{8} (same as we do) and 𝒮=r​ℬ\mathcal{S}=r\mathcal{B}. However, this seems to only be possible for quantizing weights and not activations as complexity makes runtime implementation too slow.†† † We note that QuIP# cleverly exploits symmetries of E8E_{8} to show that an R=2R=2 bit quantizer can be implemented using an LUT of size 2d​R2=282^{d\frac{R}{2}}=2^{8}, but we believe this is still too slow, and furthermore does not naturally extend to different quantization rates.

Using int8-multipliers: One often mentioned advantage of uniform quantization compared to other approaches is the fact that it approximates any matrix as a product of diagonal matrix (of norms) and an integer matrix. Thus, during multiplication one can leverage faster int-cores rather than floating-point multiplication. Note that if there exists a scaling coefficient α>0\alpha>0 such that α​Λ⊂ℤd\alpha\Lambda\subset\mathbb{Z}^{d}, then one can still use int-multipliers even for lattice-quantized vectors.

Voronoi codes/nested lattice codes: In Voronoi codes [1] the same lattice Λ\Lambda is used for both quantization and shaping. In particular, the shaping region is taken as 𝒮=q​𝒱Λ\mathcal{S}=q\mathcal{V}_{\Lambda}, where q=2Rq=2^{R} is an integer. As elaborated below, if QΛ​(x)Q_{\Lambda}(x) admits an efficient implementation, one can efficiently perform encoding and decoding to the codebook 𝒞=Λ∩(2R​𝒱Λ)≅Λ/2R​Λ\mathcal{C}=\Lambda\cap(2^{R}\mathcal{V}_{\Lambda})\cong\Lambda/2^{R}\Lambda. Moreover, in stark contrast to ball-based shaping, the encoding and decoding complexity does not depend on RR.

In summary, a good choice of lattice Λ\Lambda should therefore have: 1) efficient lattice decoding algorithm; 2) small NSM; 3) large μ⁡(𝒱Λ)\mu(\mathcal{V}_{\Lambda}); 4) be a subset of standard integer lattice ℤd\mathbb{Z}^{d}.

In this work, we use the Gosset lattice (E8E_{8}) that satisfies all these properties. It has a fast decoding algorithm (Algorithm 5), its NSM is ≈0.0716821≈1.2243​12​π​e\approx 0.0716821\approx 1.2243\frac{1}{2\pi e} [24], and its Gaussian mass μ⁡(r​𝒱E8)\mu(r\mathcal{V}_{E_{8}}) is very close to μ⁡(r​ℬ)\mu(r\mathcal{B}) (the Gosset lattice has unit covolume, so vol⁡(r​𝒱E8)=vol⁡(r​ℬ)\mathrm{vol}(r\mathcal{V}_{E_{8}})=\mathrm{vol}(r\mathcal{B})). The last point is illustrated in Figure 5, where the large loss for cubic shaping with respect to lattice shaping is also evident. The gap between Voronoi/Ball shaping and cubic shaping becomes much more significant as the dimension increases. This follows since for large dd we have that for X∼μ=μdX\sim\mu=\mu_{d} the ℓ∞\ell_{\infty} norm ‖X‖∞\|X\|_{\infty} concentrates around 2​ln⁡d\sqrt{2\ln{d}}. Thus, for r<2​2​ln⁡dr<2\sqrt{2\ln{d}} we have that μ⁡(r​CUBE)→0\mu(r\mathrm{CUBE})\to 0, whereas for any r>dreff​(1)=2​π​e​(1+o⁡(1))r>\frac{\sqrt{d}}{r_{\text{eff}}(1)}=\sqrt{2\pi e}(1+o(1)) we have that μ⁡(r​ℬ)→1\mu(r\mathcal{B})\to 1. Note that SpinQuant uses high-dimensional cubic shaping, and therefore its MSE distortion suffers a O⁡(ln⁡d)O(\ln d) multiplicative gap with respect to the optimal distortion.

Overload avoidance via union of Voronoi codes: Because we rely on lattice quantizers of relatively small dimension (d=8d=8), even if μ⁡(r​𝒱Λ)\mu(r\mathcal{V}_{\Lambda}) is very close to μ⁡(r​ℬ)\mu(r\mathcal{B}), overload events are unavoidable. This follows because in small dimension the norm of a iid Gaussian vector is not sufficiently concentrated. Thus, if one is restricted to 𝒞=β⁡(Λ∩(2R​𝒱Λ))\mathcal{C}=\beta(\Lambda\cap(2^{R}\mathcal{V}_{\Lambda})) the parameter β\beta must be taken quite large in order to keep the overload probability small. This in turn, incurs a significant penalty in the obtained distortion. As a remedy, rather than using a Voronoi code, we take 𝒞\mathcal{C} as a union of (a small number) of Voronoi codes in different scales. Namely, we take 𝒞=∪t=1kβt(Λ∩(2R𝒱Λ))\mathcal{C}=\cup_{t=1}^{k}\beta_{t}(\Lambda\cap(2^{R}\mathcal{V}_{\Lambda})), where β1<⋯<βk\beta_{1}<\cdots<\beta_{k}. The smallest values of βt\beta_{t} are set such that overload is not too common but not extremely rare, such that for most realizations of a Gaussian vector X∈ℝdX\in\mathbb{R}^{d} the distortion is close to the fundamental limit D⁡(R)D(R). Whenever XX is atypically large, there will be overload in βt​(Λ∩(2R​𝒱Λ))\beta_{t}(\Lambda\cap(2^{R}\mathcal{V}_{\Lambda})) for small tt, but not for large tt, such that the quantization error will be in βt​𝒱Λ\beta_{t}\mathcal{V}_{\Lambda} for one of the larger values of {βt}\{\beta_{t}\}.

The details of choosing kk and values of βt\beta_{t} are described in Section F. Here we only note that the overall compression rate becomes R+1d​log2​kR+{\frac{1}{d}}{\log_{2}k}. In some cases, we are using nvcomp [25] for compressing a vector of n/8n/8 chosen betas, in which case rate penalty is reduced below 1d​log2​k{\frac{1}{d}}\log_{2}k. We note that in our comparisons, including Table 1, we always use this effective rate for a fair comparison with other algorithms.

Figure 3: RMSE for quantized matrix multiplication for iid 𝒩⁡(0,1)\mathcal{N}(0,1) matrices. NestQuant algo is optimized over qq and multiple β\beta’s. Also shown is information-theoretic lower bound from (1).

NestQuant, SpinQuant and theory: As mentioned above, the use of nested lattice codes is rooted in theory. In (8) it was shown that nested lattice quantizers of high-dimensions attain the optimal rate-distortion tradeoff for matrix multiplication. Since the lattices used for proving that result do not admit efficient lattice decoding, here we resort to nn-dimensional lattices constructed as the Cartesian product of n/dn/d copies of the Gosset lattice, whose dimension is d=8d=8. To understand how much loss in efficiency this leads to, Fig. 3 compares NestQuant, SpinQuant (uniform quantization with cubic shaping) and information-theoretic lower bound (1), for synthetic data. Details of this experiment can be found in Section 5.1. We can see that NestQuant is reasonably close to the fundamental limit and significantly outperforms SpinQuant.

4 Detailed Method

4.1 Nested lattice codebook

In this section, we describe the construction for a Vector Quantization (VQ) codebook of size qdq^{d} for quantizing a dd-dimensional vector, where qq is an integer parameter. This construction is based on Voronoi codes [1] and admits efficient encoding and decoding algorithms, whenever the base lattice has an efficient closest lattice vector algorithm. Another appealing feature of Voronoi codes is that the encoding/decoding complexity is independent of the quantozation rate R=log2⁡(q)R=\log_{2}(q).

Definition 4.1 (Voronoi code [1]).

The Voronoi codebook with base lattice Λ⊂ℝd\Lambda\subset\mathbb{R}^{d} and nesting ratio q∈ℕq\in\mathbb{N}, corresponding to rate R=log2⁡(q)R=\log_{2}(q) bits/entry, is defined as C=Λ∩𝒱q​Λ=Λ∩(q​𝒱Λ)⊂ℝdC=\Lambda\cap\mathcal{V}_{q\Lambda}=\Lambda\cap(q\mathcal{V}_{\Lambda})\subset\mathbb{R}^{d}. In particular λ∈Λ\lambda\in\Lambda belongs to codebook CC iff λ∈𝒱q​Λ\lambda\in\mathcal{V}_{q\Lambda}, and |C|=qd|C|=q^{d}.

The Voronoi code consists of the minimum energy representative of each coset in Λ/q​Λ≅(ℤ/q​ℤ)d\Lambda/q\Lambda\cong(\mathbb{Z}/q\mathbb{Z})^{d}. Consequently, we can represent each coset, and hence, each codeword in CC, as an element of ℤqd\mathbb{Z}_{q}^{d} [1].

Assuming we have access to an oracle QΛ​(x)Q_{\Lambda}(x), which maps x∈ℝdx\in\mathbb{R}^{d} to its closest point in Λ\Lambda, the quantization (encoding) and dequantization (decoding) of a Voronoi code are described in Algorithm 1 and Algorithm 2, respectively. Here, G∈ℝd×dG\in\mathbb{R}^{d\times d} is a generator matrix of Λ\Lambda. The encoder first maps x∈ℝdx\in\mathbb{R}^{d} to its nearest lattice point p=QΛ​(x)p=Q_{\Lambda}(x). Since it only has a budget of d​RdR bits for describing pp, it only describes the coset of Λ/q​Λ\Lambda/q\Lambda it belongs to. This is done by sending v​mod​qv\;\mathrm{mod}\ q, where v∈ℤdv\in\mathbb{Z}^{d} is the integer vector for which p=G​vp=Gv, referred to as the coordinates of pp. Upon receiving v​mod​qv\;\mathrm{mod}\ q, the decoder knows that QΛ​(x)∈p+q​ΛQ_{\Lambda}(x)\in p+q\Lambda, and has to choose one point in this coset as the reconstruction x^∈ℝd\hat{x}\in\mathbb{R}^{d}. It chooses x^\hat{x} as the minimum energy vector in p+q​Λp+q\Lambda, corresponding to the unique point in p+q​Λ∩𝒱q​Λp+q\Lambda\cap\mathcal{V}_{q\Lambda}. We have that x^=QΛ​(x)\hat{x}=Q_{\Lambda}(x) iff QΛ​(x)∈𝒱q​Λ=q​𝒱ΛQ_{\Lambda}(x)\in\mathcal{V}_{q\Lambda}=q\mathcal{V}_{\Lambda}. When QΛ​(x)∉q​𝒱ΛQ_{\Lambda}(x)\notin q\mathcal{V}_{\Lambda} the quantizer is in overload.

Algorithm 1 Encode
Input: x∈ℝdx\in\mathbb{R}^{d}, QΛQ_{\Lambda}
p←QΛ​(x)p\leftarrow Q_{\Lambda}(x)
v←G−1​pv\leftarrow G^{-1}p ⊳\triangleright coordinates of pp
return v​mod​qv\;\mathrm{mod}\ q ⊳\triangleright quantized representation of pp
Algorithm 2 Decode
Input: c∈ℤqdc\in\mathbb{Z}_{q}^{d}, QΛQ_{\Lambda}
p←G​cp\leftarrow Gc ⊳\triangleright equivalent to answer modulo q​Λq\Lambda
return p−q​QΛ​(pq)p-q\,Q_{\Lambda}\!\bigl(\tfrac{p}{q}\bigr)

In our experiments for this paper, we used the Gosset (E8E_{8}) lattice as Λ\Lambda with d=8d=8. This lattice is a union of D8D_{8} and D8+12D_{8}+\frac{1}{2}, where D8D_{8} contains elements of ℤ8\mathbb{Z}^{8} with even sum of coordinates. There is a simple algorithm for finding the closest point in the Gosset lattice, first described in [26]. We provide the pseudocode for this algorithm together with the estimation of its runtime in Appendix C.

4.2 Matrix quantization

When quantizing a matrix, we normalize its rows, and quantize each block of dd entries using the codebook. The algorithm 3 describes the quantization procedure for each row of the matrix.

Algorithm 3 NestQuant
Input: AA — a vector of size n=d​bn=db, qq, array of kk scaling coefficients β1,…,βk\beta_{1},\ldots,\beta_{k}
Q​AQA — nn integers in {0,1​…,q−1}\{0,1\ldots,q-1\}⊳\triangleright quantized representation
BB — bb integers in {1,…,k}\{1,\ldots,k\} ⊳\triangleright scaling coefficient indices
s←∥Ai∥2s\leftarrow\lVert A_{i}\rVert_{2} ⊳\triangleright normalization coefficient
A←A​nsA\leftarrow\frac{A\sqrt{n}}{s}
for j=0j=0 to b−1b-1 do
  e​r​r=∞err=\infty
  for p=1p=1 to kk do
   v←A[dj+1..dj+d]v\leftarrow A[dj+1..dj+d]
   e​n​c←Encode​(vβp)enc\leftarrow\text{Encode}\left(\frac{v}{\beta_{p}}\right)
   r​e​c​o​n←Decode​(e​n​c)⋅βprecon\leftarrow\text{Decode}(enc)\cdot\beta_{p}
   if e​r​r>|r​e​c​o​n−v|22err>|recon-v|_{2}^{2} then
     e​r​r←|r​e​c​o​n−v|22err\leftarrow|recon-v|_{2}^{2}
     QA[dj+1..dj+d]←encQA[dj+1..dj+d]\leftarrow enc
     Bj←pB_{j}\leftarrow p
   end if
  end for
end for
Output: Q​AQA, BB, ss

We can take dot products of quantized vectors without complete dequantization using algorithm 4. We use it in the generation stage on linear layers and for querying the KV cache.

Algorithm 4 Dot product
Input: Q​A1QA_{1}, B1B_{1}, s1s_{1} and Q​A2QA_{2}, B2B_{2}, s2s_{2} — representations of two vectors of size n=d​bn=db from Algorithm 3, array β\beta
a​n​s←0ans\leftarrow 0
for j=0j=0 to b−1b-1 do
  p1←Decode(QA1[dj+1..dj+d])p_{1}\leftarrow\text{Decode}(QA_{1}[dj+1..dj+d])
  p2←Decode(QA2[dj+1..dj+d])p_{2}\leftarrow\text{Decode}(QA_{2}[dj+1..dj+d])
  a​n​s←a​n​s+(p1⋅p2)​βB1​[j]​βB2​[j]ans\leftarrow ans+(p_{1}\cdot p_{2})\beta_{B_{1}[j]}\beta_{B_{2}[j]}
end for
return a​n​sans

4.3 LLM quantization

Figure 4: The quantization scheme of multi-head attention. HH is Hadamard rotation described in 4.3. 𝒬\mathcal{Q} is the quantization function described in 4.2

Recall that we apply a rotation matrix HH to every weight-activation pair of a linear layer without changing the output of the network. Let nn be the number of input features to the layer. Following [18, 16, 27]:

  • •

    If n=2kn=2^{k}, we set HH to be Hadamard matrix obtained by Sylvester’s construction

  • •

    Otherwise, we decompose n=2k​mn=2^{k}m, such that mm is small and there exists a Hadamard matrix H1H_{1} of size mm. We construct Hadamard matrix H2H_{2} of size 2k2^{k} using Sylvester’s construction, and set U=H1⊗H2U=H_{1}\otimes H_{2}.

Note that it’s possible to multiply an r×nr\times n matrix by HH in O⁡(r​n​log⁡n)O(rn\log n) in the first case and O⁡(r​n​(log⁡n+m))O(rn(\log n+m)) in the second case, which is negligible to other computational costs and can be done online.

In NestQuant, we quantize all weights, activations, keys, and values using Algorithm 3. We merge the Hadamard rotation with the weights and quantize them. We also apply the Hadamard rotation and quantization to the activations before linear layers. We also apply rotation to keys and queries, because it will not change the attention scores, and we quantize keys and values before putting them in the KV cache. Figure 4 illustrates the procedure for multi-head attention layers.

When quantizing a weight, we modify the NestQuant algorithm by introducing corrections to unquantized weights when a certain vector piece is quantized via the QA-LDLQ mechanism, described in section 4.5. It is based on LDLQ algorithm, described in section 4.1 of [18].

4.4 Optimal scaling coefficients

One of the important parts of the algorithm is finding the optimal set of βi\beta_{i}. Given the distribution of dd-dimensional vectors that are quantized via a Voronoi codebook, it is possible to find an optimal set of given size exactly using a dynamic programming approach, which is described in Appendix F.

4.5 LDLQ with quantized activations (QA-LDLQ)

We note that in the presence of activation quantization, the LDLQ and GPTQ algorithms for weight quantization become significantly suboptimal ‡‡ ‡ In fact, using original LDLQ on Llama-3-70B produces ∞\infty perplexity at W4A4 setting due to significant outliers in layer 0., thus necessitating a correction (QA-LDLQ) that we describe here.

Let WW be the weight (shape a×na\times n), XX be a random original (unquantized) activation vector (shape n×1n\times 1), and ZZ be XX’s quantization error, which we model as random zero-mean noise (shape n×1n\times 1) independent of XX. Then, if UU is the quantized weight, the output quantization error becomes δ⁡(U):=W​X−U⁡(X+Z)\delta(U):=WX-U(X+Z), so that the loss to be minimized is 𝔼⁡[‖δ⁡(U)‖2]\mathbb{E}[\|\delta(U)\|^{2}] instead of 𝔼⁡[‖(W−U)​X‖2]\mathbb{E}[\|(W-U)X\|^{2}] that LDLQ and GPTQ minimize.

Lemma 4.2.

Suppose ZZ is independent from XX, 𝔼⁡[Z]=0\mathbb{E}[Z]=0, and let H=𝔼⁡[X​X⊤]H=\mathbb{E}[XX^{\top}] and J=𝔼⁡[Z​Z⊤]J=\mathbb{E}[ZZ^{\top}]. Then for any set 𝒞Q⊂ℝa×n\mathcal{C}_{Q}\subset\mathbb{R}^{a\times n}

U∗\displaystyle U^{*} =arg⁡minU∈𝒞Q⁡𝔼​[‖δ⁡(U)‖2]\displaystyle=\argmin_{U\in\mathcal{C}_{Q}}\mathbb{E}[\|\delta(U)\|^{2}]
=arg⁡minU∈𝒞Q⁡(W~−U)​(H+J)​(W~−U)⊤,\displaystyle=\argmin_{U\in\mathcal{C}_{Q}}(\tilde{W}-U)(H+J)(\tilde{W}-U)^{\top}, (6)

where W~=W​H​(H+J)−1\tilde{W}=WH(H+J)^{-1}.

Thus, QA-LDLQ consists of a) computing W~\tilde{W}; b) running the standard LDLQ but with matrix W~\tilde{W} as input and Hessian set to H+JH+J. More details, practical aspects of QA-LDLQ, as well as proof of the lemma are found in Appendix B.

4.6 Algorithm summary

Here we describe the main steps of NestQuant.

  1. 1.

    Collect the statistics for LDLQ via calibration data. For each linear layer with in-dimension dd, we compute a d×dd\times d “Hessian” matrix HH.

  2. 2.

    We choose an initial set of scaling coefficients β^\hat{\beta}, and for each weight we simulate LDLQ quantization with these coefficients, getting a set of 8-dimensional vectors to quantize.

  3. 3.

    We run a dynamic programming algorithm described in Appendix F on the 8-vectors to find the optimal β\beta-values for each weight matrix.

  4. 4.

    We also run the dynamic programming algorithm for activations, keys, and values for each layer. To get the distribution of 8-vectors, we run the model on a small set of examples.

  5. 5.

    We quantize the weights using QA-LDLQ and precomputed β\beta.

  6. 6.

    During inference, we quantize any activation before it is passed to the linear layer, and any KV cache entry before it is saved.

Note that we do not undertake any expensive (but surely useful) fine-tuning, such as optimizing rotation matrices or post-quantization training, as in [16] and [18], since our goal is demonstrating the basic primitive, not obtaining the absolute SOTA numbers.

5 Experiments

5.1 Simulated Data

We compared the mean L2L_{2} loss per entry of SpinQuant to the uniform L∞L_{\infty}-scaling quantizer (used in SpinQuant and other methods). The mean L2L_{2} loss per entry for the product of two matrices A∈ℝn×k,B∈ℝm×kA\in\mathbb{R}^{n\times k},B\in\mathbb{R}^{m\times k} is computed as ∥A​BT−A^​B^T∥2n​m\frac{\lVert AB^{T}-\hat{A}\hat{B}^{T}\rVert_{2}}{nm}. We set n=k=m=4096n=k=m=4096 and sampled two matrices A,BA,B with unit normal distribution Ai​j,Bi​j∼𝒩⁡(0,1)A_{ij},B_{ij}\sim\mathcal{N}(0,1). We compare to the lower bound from (1).

For NestQuant, we do a grid search over (q,k)(q,k). For given qq and kk, we find the best subset in 12⋅{1,2,…,50}\frac{1}{2}\cdot\{1,2,\ldots,50\} of scaling coefficients β\beta of size kk using the algorithm from Appendix F. Then we calculate the expected bits per entry computed as log2⁡q+18​∑i=1kp⁡(βi)​log2⁡p⁡(βi)\log_{2}{q}+\frac{1}{8}\sum_{i=1}^{k}p(\beta_{i})\log_{2}p(\beta_{i}) where p⁡(βi)p(\beta_{i}) is the probability that the ii’th beta is used in quantization. In Figure 3, we plot the efficient frontier of bits per entry vs root mean L2L_{2} loss.

5.2 Llama results

Bits (W-A-KV) Method Llama-2-7B Llama-2-13B Llama-2-70B Llama-3-8B Llama-3-70B 16-16-16 Floating point 5.47 4.88 3.32 6.14 2.86 4-16-16 QuaRot 5.60 5.00 3.41 - - QuIP# 5.56 4.95 3.38 - - OstQuant 5.64 4.94 3.41 6.53 3.19 NestQuant 5.53 4.93 3.38 6.31 3.14 NestQuantM 5.55 4.95 - 6.35 - 4-16-4 NestQuant 5.57 4.96 3.39 6.37 3.19 NestQuantM 5.59 4.99 - 6.49 - 4-4-16 SpinQuant 5.9 5.2 3.8 7.1 - OstQuant 5.60 5.14 3.57 7.24 3.97 DuQuant 6.08 5.33 3.76 8.06 - 4-4-4 QuaRot 6.10 5.40 3.79 8.16 6.66 SpinQuant 5.9 5.3 3.8 7.3 - OstQuant 5.91 5.25 3.59 7.29 4.01 NestQuant 5.67 5.03 3.49 6.63 3.61 NestQuantM 5.73 5.07 - 6.82 -

Table 2: The wikitext2 perplexity at context window of 2048 for various quantization methods of Llama models.

We quantize the Llama-3-8B model [28] using different values of qq. We choose the number of scaling coefficients (kk) to be equal to 44, the Section H.1 explains the rationale behind this choice. More details on the hyperparameter choice of the experiments are in Appendix G. For each experiment, we compute the number of bits per entry similar to Section 5.1, but for the setup of compressed β\beta indices, we run the zstd compression algorithm instead of using the entropy of the distribution. As our evaluation metric, we use the perplexity on the validation split of wikitext2 with context length 20482048.

We also perform the evaluation of NestQuant on various zero-shot benchmarks: ARC-Easy and ARC-Challenge [29], Hellaswag [30], [31], and Winogrande [32]. The results on 4-bit models with comparisons to other models are summarized in Table 1.

q Bits Bits (no zstd) W W + KV W + KV + A
14 3.99 4.06 6.308 6.379 6.633
12 3.76 3.83 6.376 6.475 6.841
10 3.50 3.57 6.486 6.640 7.251
8 3.18 3.25 6.700 6.968 7.989
Table 3: Wikitext2 perplexity of NestQuant quantization of Llama-3-8B at different rates. The "bits" column is the bit rate per entry with zstd compression of scaling coefficients, and "bits (no zstd)" is the bit rate without compression. The "W", "W+KV", and "W+KV+A" describe the quantization regime (whether weights, KV cache, or activations are quantized). The perplexity of non-quantized model is 6.1396.139

5.3 Comparison to other methods

In comparisons to other methods, we focus on 4-bit setup, choosing q=14q=14 and k=4k=4. We show the WikiText2 perplexity comparisons for multiple Llama models in table 2, and other benchmark comparisons (for Llama-3-8B) in table 1. The methods that we include in the table are SpinQuant [16], QuIP# [18], QuaRot [15], DuQuant [27], and OstQuant [33]. On the WikiText2 dataset, we computed the perplexity scores of the quantized models on context size 2048.

NestQuant consistently achieves better perplexity metrics across different models for both weight-only regime and full quantization. In fact, for all models we have tested, except Llama-2-7B, NestQuant with W4A4KV4 (4-bit weights, activations, and KV-cache) quantization even outperforms previous works with W4A4KV16. On W4KV4A4 (4-bit weights, KV-cache, and activations) quantization of Llama 3-8B we achieve a perplexity score of 6.6 compared to  7.3 in SpinQuant and OstQuant. Even without LDLQ, we achieve a perplexity score of 6.8, which is still better. Finally, NestQuant outperforms QuIP# for weight-only Llama-2 quantization.

In addition, we evaluate a simpler version of the algorithm that is easier to run in hardware that is called NestQuantM. More detailes on it can be found in section D.

We present ablation studies for our design choices in in Appendix H.

Impact statement

This paper presents work whose goal is to advance the field of Machine Learning. There are many potential societal consequences of our work, none which we feel must be specifically highlighted here.

Acknowledgments

We want to thank Nikita Lazarev and Shang Yang for important discussions that helped us to implement an efficient CUDA kernel, as well as Guangxuan Xiao for discussions about the algorithm.

This work was supported in part by the National Science Foundation under Grant No CCF-2131115 and the ISF under Grant 1641/21. We thank Foundry.ai and Lambda Labs for compute resources.

Research was supported, in part, by the United States Air Force Research Laboratory and the United States Air Force Artificial Intelligence Accelerator under Cooperative Agreement Number FA8750-19-2-1000. The views and conclusions contained in this document are those of the authors and should not be interpreted as representing the official policies, either expressed or implied, of the United States Air Force or the U.S. Government. The U.S. Government is authorized to reproduce and distribute reprints for Government purposes notwithstanding any copyright notation herein.

References

  • [1] J. Conway and N. Sloane. A fast encoding method for lattice codes and quantizers. IEEE Transactions on Information Theory, 29(6):820–824, 1983.
  • [2] Y. Polyanskiy and Y. Wu. Information Theory: From Coding to Learning. Cambridge University Press, 2024.
  • [3] R. Zamir. Lattice Coding for Signals and Networks: A Structured Coding Approach to Quantization, Modulation, and Multiuser Information Theory. Cambridge University Press, 2014.
  • [4] A. J. Stam. Limit theorems for uniform distributions on spheres in high-dimensional euclidean spaces. Journal of Applied Probability, 19(1):221–228, 1982.
  • [5] A. Gersho. Asymptotically optimal block quantization. IEEE Transactions on Information Theory, 25(4):373–380, 1979.
  • [6] A. C. Hung and T. H. Meng. Multidimensional rotations for robust quantization of image data. IEEE Transactions on Image Processing, 7(1):1–12, 1998.
  • [7] R. Hadad and U. Erez. Dithered quantization via orthogonal transformations. IEEE Transactions on Signal Processing, 64(22):5887–5900, 2016.
  • [8] O. Ordentlich and Y. Polyanskiy. Optimal quantization for matrix multiplication, 2024.
  • [9] R. Zamir and M. Feder. On lattice quantization noise. IEEE Transactions on Information Theory, 42(4):1152–1159, 1996.
  • [10] M. S. Charikar. Similarity estimation techniques from rounding algorithms. In Proceedings of the 34th Annual ACM Symposium on Theory of Computing (STOC 2002), pages 380–388, 2002.
  • [11] M. Datar, N. Immorlica, P. Indyk, and V. S. Mirrokni. Locality-sensitive hashing scheme based on pp-stable distributions. In Proceedings of the 20th Annual Symposium on Computational Geometry (SCG 2004), pages 253–262, 2004.
  • [12] A. Andoni and P. Indyk. Near-optimal hashing algorithms for approximate nearest neighbor in high dimensions. Communications of the ACM, 51(1):117–122, 2008.
  • [13] T. Dettmers, M. Lewis, Y. Belkada, and L. Zettlemoyer. LLM.int8(): 8-bit matrix multiplication for transformers at scale. In Advances in Neural Information Processing Systems 35 (NeurIPS 2022), pages 29498–29512, 2022.
  • [14] G. Xiao, J. Lin, M. Seznec, H. Wu, J. Demouth, and S. Han. Smoothquant: Accurate and efficient post-training quantization for large language models. In Proceedings of the 40th International Conference on Machine Learning (ICML 2023), volume 202 of Proceedings of Machine Learning Research, pages 3732–3754, 2023.
  • [15] S. Ashkboos, A. Mohtashami, M. L. Croci, B. Li, P. Cameron, M. Jaggi, D. Alistarh, T. Hoefler, and J. Hensman. Quarot: Outlier-free 4-bit inference in rotated llms. In Advances in Neural Information Processing Systems 38 (NeurIPS 2024), 2024.
  • [16] Z. Liu, C. Zhao, I. Fedorov, B. Soran, D. Choudhary, R. Krishnamoorthi, V. Chandra, Y. Tian, and T. Blankevoort. Spinquant: Llm quantization with learned rotations, 2024.
  • [17] J. Li, F. Li, and S. Todorovic. Efficient riemannian optimization on the stiefel manifold via the cayley transform. In International Conference on Learning Representations (ICLR 2020), 2020.
  • [18] A. Tseng, J. Chee, Q. Sun, V. Kuleshov, and C. De Sa. Quip#: Even better llm quantization with hadamard incoherence and lattice codebooks. In Proceedings of the 41st International Conference on Machine Learning (ICML 2024), volume 235 of Proceedings of Machine Learning Research, pages 48630–48656, 2024.
  • [19] B.-H. Juang and A. H. Gray. Multiple stage vector quantization for speech coding. In IEEE International Conference on Acoustics, Speech, and Signal Processing, 1982.
  • [20] A. Tseng, Q. Sun, D. Hou, and C. De Sa. Qtip: Quantization with trellises and incoherence processing. In Advances in Neural Information Processing Systems 38 (NeurIPS 2024), 2024.
  • [21] J. Chee, Y. Cai, V. Kuleshov, and C. De Sa. Quip: 2-bit quantization of large language models with guarantees. In Advances in Neural Information Processing Systems 37 (NeurIPS 2023), 2023.
  • [22] E. Frantar, S. P. Singh, and D. Alistarh. Optimal brain compression: A framework for accurate post-training quantization and pruning. In Advances in Neural Information Processing Systems 36 (NeurIPS 2022), 2022.
  • [23] I. Dinur, G. Kindler, R. Raz, and S. Safra. Approximating CVP to within almost-polynomial factors is np-hard. Combinatorica, 23(2):205–243, 2003.
  • [24] E. Agrell and B. Allen. On the best lattice quantizers. IEEE Transactions on Information Theory, 2023.
  • [25] NVIDIA Corporation. nvcomp: Gpu-accelerated compression library, 2025. Accessed 2025-01-30.
  • [26] J. Conway and N. Sloane. Fast quantizing and decoding algorithms for lattice quantizers and codes. IEEE Transactions on Information Theory, 28(2):227–232, 1982.
  • [27] H. Lin, H. Xu, Y. Wu, J. Cui, Y. Zhang, L. Mou, L. Song, Z. Sun, and Y. Wei. Duquant: Distributing outliers via dual transformation makes stronger quantized llms. In Advances in Neural Information Processing Systems 38 (NeurIPS 2024), 2024.
  • [28] Llama Team, AI @ Meta. The llama 3 herd of models, 2024.
  • [29] P. Clark, I. Cowhey, O. Etzioni, T. Khot, A. Sabharwal, C. Schoenick, and O. Tafjord. Think you have solved question answering? try arc, the ai2 reasoning challenge, 2018.
  • [30] R. Zellers, A. Holtzman, Y. Bisk, A. Farhadi, and Y. Choi. Hellaswag: Can a machine really finish your sentence? In Proceedings of the 57th Annual Meeting of the Association for Computational Linguistics (ACL 2019), pages 4791–4800, 2019.
  • [31] Y. Bisk, R. Zellers, R. Le Bras, J. Gao, and Y. Choi. Piqa: Reasoning about physical commonsense in natural language. In Proceedings of the 34th AAAI Conference on Artificial Intelligence (AAAI 2020), pages 7432–7439, 2020.
  • [32] K. Sakaguchi, R. Le Bras, C. Bhagavatula, and Y. Choi. Winogrande: An adversarial winograd schema challenge at scale. In Proceedings of the 34th AAAI Conference on Artificial Intelligence (AAAI 2020), pages 8732–8740, 2020.
  • [33] X. Hu, Y. Cheng, D. Yang, Z. Xu, Z. Yuan, J. Yu, C. Xu, Z. Jiang, and S. Zhou. Ostquant: Refining large language model quantization with orthogonal and scaling transformations for better distribution fitting. In International Conference on Learning Representations (ICLR 2025), 2025.
  • [34] E. Frantar, S. Ashkboos, T. Hoefler, and D. Alistarh. Gptq: Accurate post-training quantization for generative pre-trained transformers. In Advances in Neural Information Processing Systems 37 (NeurIPS 2023), 2023.
  • [35] S. Merity, C. Xiong, J. Bradbury, and R. Socher. Pointer sentinel mixture models. In International Conference on Learning Representations (ICLR 2017), 2017.

Appendix A Figures

Refer to caption
Figure 5: Complement Gaussian measure of a 88-dimensional cube (corresponding to shaping using an ℓ∞\ell_{\infty} ball), a Voronoi region of the Gosset lattice E8E_{8} (corresponding to shaping using Voronoi codes with base lattice E8E_{8}), and a Euclidean ball (corresponding to shaping with a ball, which does not admit efficient implementation)

Appendix B Details of QA-LDLQ

LDLQ details. We briefly mention the details of LDLQ (for more details we refer the reader to [34] or [21]). The main observation is that we want to minimize ∥(W−U)​X∥2\lVert(W-U)X\rVert^{2} where UU is the quantized weight and XX is the matrix of activations. We observe that ∥(W−U)​X∥2=∥(W−U)​LX​DX∥2\lVert(W-U)X\rVert^{2}=\lVert(W-U)L_{X}\sqrt{D_{X}}\rVert^{2} where LX​DX​LX⊤=H=𝔼⁡[X​X⊤]L_{X}D_{X}L_{X}^{\top}=H=\mathbb{E}[XX^{\top}] and LXL_{X} is a uni-lower-triangular matrix and DXD_{X} is diagonal. LDLQ proceeds by optimally quantizing the last column of UU (for example using the quantizer for NestQuant). It then recursively removes the feedback ii, (W−U)​(LX−I)​Ei(W-U)(L_{X}-I)E_{i} from the rest of the columns (here EiE_{i} is the zero matrix with the i​iii index set to 1). This may be written compactly as U=𝒬⁡(W−(W−U)​(LX−I))U=\mathcal{Q}(W-(W-U)(L_{X}-I)) where 𝒬\mathcal{Q} quantizes each column separately.

Motivation. When quantizing Llama-3-70B with original LDLQ, we have noticed very poor (ppl ∼104\sim 10^{4}) performance. We have observed that the reason for this is quantization of activations for a small subset of linear layers (4 out of 560), and if they are left in full precision, the perplexity becomes more reasonable (less than 4). We have found them by choosing the layers with the largest amplification ratio — a concept we define next.

Consider a weight WW (shape a×na\times n), where nn is the number of in-features, and aa is the number of out-features. We define amplification of a random vector XX (OPENn×1)n\times 1) by WW as α⁡(W,X):=𝔼⁡[‖W​X‖2]𝔼⁡[‖X‖2]\alpha(W,X):=\frac{\mathbb{E}[\left\lVert WX\right\rVert_{2}]}{\mathbb{E}[\left\lVert X\right\rVert_{2}]}. Then, if XX is the distribution of input activations, and ZZ is a random Gaussian vector, we define the amplification ratio for WW as α⁡(W,Z)α⁡(W,X)\frac{\alpha(W,Z)}{\alpha(W,X)}. We note that if a layer has large amplification ratio, its activations are harder to quantize, because the magnitude of quantization noise will be increased much more significantly than the magnitude of the activations themselves. One extreme example of this issue is the value projection of the attention of the first transformer block in Llama-3-70B. This layer has an amplification ratio of ∼157\sim 157, as computed from 1010 wikitext2 sequences of length 20482048. This makes naive 44-bit quantization of activations nearly impossible, as even small perturbations of the input of the layer are greatly amplified in the output.

We have developed QA-LDLQ to mitigate the issue of large amplification ratio. We model quantization noise as a random Gaussian vector with mean 00 and covariance matrix J=ε2​IJ=\varepsilon^{2}I, where ε2\varepsilon^{2} depends on the quantization rate and the statistics of XX. For large values of ε2\varepsilon^{2} the modified weight matrix W~=W​H​(H+ε2​I)−1\tilde{W}=WH(H+\varepsilon^{2}I)^{-1} is more robust to random perturbations of the input (thus, decreasing the amplification ratio), but on the other hand, large ε2\varepsilon^{2} results in a larger bias, as expressed in the term C⁡(W,H,J)C(W,H,J) below. Figure 6 demonstrates this tradeoff for the value projection layer mentioned earlier.

Figure 6: We run QA-LDLQ for value projection layer of the first transformer block of Llama-3-70B. We try different values of ε\varepsilon on logarithmic scale from 10−510^{-5} to 11. For each ε\varepsilon, we find modified weight W~\tilde{W}, and plot the amplification ratio for W~\tilde{W} in yy-axis, as well as how close the outputs of the weight W~\tilde{W} to the outputs of weight WW. The value on xx axis is defined as 1−R2:=𝔼​‖W​X−W~​X‖2Var⁡(W​X)1-R^{2}:=\frac{\mathbb{E}\|WX-\tilde{W}X\|^{2}}{\mathrm{Var}(WX)}, where XX contains activation inputs from 1010 sequences of length 20482048 from wikitext2. The right plot is bottom right corner of the left plot, zoomed in. We note that by paying a small price in the accuracy of the weight, we can reduce the amplification ratio dramatically.
Proof of Lemma 4.2.

Recalling the definition of W~=W​H​(H+J)−1\tilde{W}=WH(H+J)^{-1}, that XX and ZZ are statistically independent, and ZZ has zero mean, and that H,JH,J are positive semi-definite symmetric matrices, we have

𝔼​‖δ⁡(U)‖2\displaystyle\mathbb{E}\|\delta(U)\|^{2} =𝔼​‖(W−U)​X−U​Z‖2=(W−U)​H​(W−U)⊤+U​J​U⊤\displaystyle=\mathbb{E}\|(W-U)X-UZ\|^{2}=(W-U)H(W-U)^{\top}+UJU^{\top} (7)
=(W~−U)​(H+J)​(W~−U)+C⁡(W,H,J),\displaystyle=(\tilde{W}-U)(H+J)(\tilde{W}-U)+C(W,H,J), (8)

where

C⁡(W,H,J)=W⁡(H−H​(H+J)−1​H)​W⊤,\displaystyle C(W,H,J)=W(H-H(H+J)^{-1}H)W^{\top}, (9)

is independent of UU, and therefore does not affect the minimization. ∎

Appendix C Gosset oracle

In this section, we discuss the algorithm for finding the nearest neighbour in E8E_{8} lattice and estimate its performance in FLOPs (Floating Point Operations). We note that E8=D8∪D8+12E_{8}=D_{8}\cup D_{8}+\frac{1}{2}, where D8D_{8} contains vectors in ℤ8\mathbb{Z}_{8} with even sum of coordinates. To compute VE8​(x)V_{E_{8}}(x), we compute two candidate points: x1=VD8​(x)x_{1}=V_{D_{8}}(x) and x2=VD8+12​(x)x_{2}=V_{D_{8}+\frac{1}{2}}(x), and choose the one that has smaller L2L^{2} distance to xx.

To get VD8​(x)V_{D_{8}}(x), we can round each coordinate to the nearest integer. If the sum of rounded coordinates is odd, we need to "flip" the rounding direction of the coordinate for which the flip would cost the least. Note that finding the closest point in VD8+12V_{D_{8}+\frac{1}{2}} works the same, but the rounding grid now contains half-integers, not integers.

In algorithm 5, we first round our vector down (getting dd) and compute the mask (gg) of whether it’s optimal to round up for D8D_{8}. We note that the optimal rounding for D8+12D_{8}+\frac{1}{2} is d+0.5d+0.5, while the optimal rounding for D8D_{8} is d+gd+g.

We want to understand whether rounding to D8D_{8} or D8+12D_{8}+\frac{1}{2} is better. Let d​i​s​tidist_{i} be the absolute distance from the ii-th entry xi∈[di,di+1]x_{i}\in[d_{i},d_{i}+1] to the middle of this integer segment di+0.5=x2,id_{i}+0.5=x_{2,i}. We note that the contribution of this point to the MSE for D8D_{8} is (0.5−d​i​s​ti)2(0.5-dist_{i})^{2}, while for D8+12D_{8}+\frac{1}{2} is d​i​s​ti2dist_{i}^{2}. The difference is: 0.25−d​i​s​ti+d​i​s​ti2−d​i​s​ti2=0.25−d​i​s​ti0.25-dist_{i}+dist_{i}^{2}-dist_{i}^{2}=0.25-dist_{i}. If the sum of this value over ii is negative (i.e. ∑d​i​s​ti>2\sum dist_{i}>2), it’s optimal to quantize to D8D_{8}, otherwise to D8+12D_{8}+\frac{1}{2}. In pseudocode, we store ∑d​i​s​ti\sum dist_{i} as Δ\Delta

We note that we should check the constraint that the sum of coordinates in D8D_{8} is even, and if it is not, “flip" one of the rounding directions. The optimal coordinate to flip can be determined through d​i​s​tdist, and the new value of flipped coordinate — through gg. We also need to update Δ\Delta given that the MSE difference changes.

The full pseudocode of the algorithm is in Algorithm 5.

Algorithm 5 Oracle for the Gosset lattice
1: Input: x∈ℝ8x\in\mathbb{R}^{8}
2: d←floor​(x)d\leftarrow\text{floor}(x)
3: x2←d+0.5x_{2}\leftarrow d+0.5
4: g←(x>x2)g\leftarrow(x>x_{2})
5: s←2⋅g−1s\leftarrow 2\cdot g-1
6: x1←d+gx_{1}\leftarrow d+g
7: d​i​s​t←(x−x2)⋅sdist\leftarrow(x-x_{2})\cdot s
8: Δ←∑id​i​s​ti\Delta\leftarrow\sum_{i}dist_{i}
9: if ∑ix1,i\sum_{i}x_{1,i} is odd then
10:   p​o​s=arg⁡min⁡distpos=\argmin dist
11:   x1,p​o​s←x1,p​o​s−s1,p​o​sx_{1,pos}\leftarrow x_{1,pos}-s_{1,pos}
12:   Δ←Δ+2⋅d​i​s​tp​o​s−1\Delta\leftarrow\Delta+2\cdot dist_{pos}-1
13: end if
14: if ∑ix2,i\sum_{i}x_{2,i} is odd then
15:   p​o​s=arg⁡max⁡distpos=\argmax dist
16:   x2,p​o​s←x2,p​o​s+g2,p​o​sx_{2,pos}\leftarrow x_{2,pos}+g_{2,pos}
17:   Δ←Δ+1−2⋅d​i​s​tp​o​s\Delta\leftarrow\Delta+1-2\cdot dist_{pos}
18: end if
19: if Δ>2\Delta>2 then
20:   return x1x_{1}
21: else
22:   return x2x_{2}
23: end if

Appendix D NestQuantM algorithm

Due to the high complexity of arg⁡min\argmin and arg⁡max\argmax operations in algorithm 5 for the fast hardware implementation, we propose a different, more simple version of NestQuant decoding. Instead of using arg⁡max\argmax and arg⁡min\argmin on lines 10 and 15, we always assign p​o​spos to be equal to 11 (thus, indicating that we always "flip" the rounding direction of the first coordinate to fix the parity). Note that this change is only applied in the decoding stage, while during encoding we use the full version of the algorithm, which keeps the granular error the same. Let’s denote our modified Gosset oracle as f:ℝ8→ℝ8f\colon\mathbb{R}^{8}\to\mathbb{R}^{8}.

Lemma D.1.

For a vector x∈ℝ8x\in\mathbb{R}^{8} and a vector v∈E8v\in E_{8}, f⁡(x+v)=f⁡(x)+vf(x+v)=f(x)+v.

Proof.

Recall that D8⊂ℤ8D_{8}\subset\mathbb{Z}^{8} contains all the integer vector with even sum of coordinates. The modified Gosset oracle ff uses modified D8D_{8} oracle gg, and at input xx constructs two candidate points c1​(x)∈D8c_{1}(x)\in D_{8} and c2​(x)∈D8+12c_{2}(x)\in D_{8}+\frac{1}{2}. Then, the algorithm chooses the closest point among these candidates to xx. Note that c1​(x)=g​(x)c_{1}(x)=g(x) and c2​(x)=g⁡(x−12)+12c_{2}(x)=g\left(x-\frac{1}{2}\right)+\frac{1}{2}.

We will prove that for u∈D8u\in D_{8}, g⁡(x+u)=g⁡(x)+ug(x+u)=g(x)+u for any x∈ℝ8x\in\mathbb{R}^{8}. Now, let’s show the original lemma. Note that since we are choosing the closest candidate point, if the condition on candidate sets {c1​(x+v),c2​(x+v)}={c1​(x),c2​(x)}+v\{c_{1}(x+v),c_{2}(x+v)\}=\{c_{1}(x),c_{2}(x)\}+v holds, then f⁡(x+v)=f⁡(x)+vf(x+v)=f(x)+v. Now, consider two cases:

  1. 1.

    v∈D8v\in D_{8}. Then:

    c1​(x+v)\displaystyle c_{1}(x+v) =g⁡(x+v)=g⁡(x)+v=c1​(x)+v\displaystyle=g(x+v)=g(x)+v=c_{1}(x)+v
    c2​(x+v)\displaystyle c_{2}(x+v) =g⁡(x+v−12)+12=g⁡(x−12)+12+v=c2​(x)+v\displaystyle=g\left(x+v-\frac{1}{2}\right)+\frac{1}{2}=g\left(x-\frac{1}{2}\right)+\frac{1}{2}+v=c_{2}(x)+v
  2. 2.

    v∈D8+12v\in D_{8}+\frac{1}{2}. Then, we say that v=u−12=w+12v=u-\frac{1}{2}=w+\frac{1}{2} for u,v∈D8u,v\in D_{8}.

    c1​(x+v)\displaystyle c_{1}(x+v) =g⁡(x+v)=g⁡(x+u−12)=g⁡(x−12)+u=g⁡(x−12)+12+v=c2​(x)+v\displaystyle=g(x+v)=g\left(x+u-\frac{1}{2}\right)=g\left(x-\frac{1}{2}\right)+u=g\left(x-\frac{1}{2}\right)+\frac{1}{2}+v=c_{2}(x)+v
    c2​(x+v)\displaystyle c_{2}(x+v) =g⁡(x−12+v)+12=g⁡(x+w)+12=g⁡(x)+w+12=g⁡(x)+v=c1​(x)+v\displaystyle=g\left(x-\frac{1}{2}+v\right)+\frac{1}{2}=g(x+w)+\frac{1}{2}=g(x)+w+\frac{1}{2}=g(x)+v=c_{1}(x)+v

Thus, in both cases the condition on candidate sets holds, and we get f⁡(x+v)=f⁡(x)+vf(x+v)=f(x)+v.

Now we show that if u∈D8u\in D_{8}, g⁡(x+u)=g⁡(x)+ug(x+u)=g(x)+u. Note that when evaluating g⁡(x+u)g(x+u), we will get the same rounding directions as in g⁡(x)g(x), since uu is an integer vector. Since uu has even sum of coordinates, the parity will also be the same. Then, our decision to flip the rounding of the first coordinate will also match as well. Therefore, the vector between original and rounded coordinates will be the same:

g⁡(x)−x=g⁡(x+u)−x−u⇒g⁡(x+u)=g⁡(x)+ug(x)-x=g(x+u)-x-u\Rightarrow g(x+u)=g(x)+u

∎

Let vv be the vector we obtain after rounding to E8E_{8}, cc be the lattice coordinates of vv, and GG be the generating matrix of the lattice. The compressed representation is c​mod​qc\;\mathrm{mod}\ q, which corresponds to a point v′=G⁡(c​mod​q)v^{\prime}=G(c\;\mathrm{mod}\ q). The reconstructed point v^\hat{v} is defined to be v′−q​f​(v′/q)v^{\prime}-qf(v^{\prime}/q) by the decoding algorithm. Note that v−v′∈q​E8v-v^{\prime}\in qE_{8}. Let’s assume that f⁡(v/q)=0f(v/q)=0. Then:

v^=v′−q​f​(v′q)=v′−q​f​(vq+v′−vq)=v′−q⁡(f⁡(vq)+v′−vq)=v′−q⋅v′−vq=v\hat{v}=v^{\prime}-qf\left(\frac{v^{\prime}}{q}\right)=v^{\prime}-qf\left(\frac{v}{q}+\frac{v^{\prime}-v}{q}\right)=v^{\prime}-q\left(f\left(\frac{v}{q}\right)+\frac{v^{\prime}-v}{q}\right)=v^{\prime}-q\cdot\frac{v^{\prime}-v}{q}=v

We have used Lemma D.1 in the third equality and our assumption in the fourth equality. Given this fact, we conclude that when using NestQuantM for decoding, the composition of encoding and decoding functions (for a fixed β\beta) is similar to the original NestQuant, except the shaping region has changed to the set of points v∈E8v\in E_{8} such that f⁡(v/q)=0f(v/q)=0. Since ff is close to the original Gosset oracle, we expect this region to still capture Gaussian probability density well. In case of overload errors, we are still able to choose a larger value of beta due to multi-beta strategy.

Appendix E CUDA Kernel Implementation

The decoding algorithm 2 consists of a basis change p←G​cp\leftarrow Gc and a subsequent computation of the coset of pq\frac{p}{q},

pq−QΛ​(pq).\frac{p}{q}-Q_{\Lambda}\left(\frac{p}{q}\right).

We leverage the asymmetry between encoding and decoding choosing GG which is fast to decode. In particular we use

G=(1000000010200000100020001000002014222222100200001000020010000002)G=\begin{pmatrix}1&0&0&0&0&0&0&0\\ 1&0&2&0&0&0&0&0\\ 1&0&0&0&2&0&0&0\\ 1&0&0&0&0&0&2&0\\ 1&4&2&2&2&2&2&2\\ 1&0&0&2&0&0&0&0\\ 1&0&0&0&0&2&0&0\\ 1&0&0&0&0&0&0&2\end{pmatrix}

GG is not a basis for E8E_{8} but for 2​E82E_{8}. We use GG because we want to work in integer-lattice as half-integers cannot be represented with integers. Moreover, we represent 8-vector x→\vec{x} as two 32 bit integers each representing the two parts of the vector x0,x1,x2,x3x_{0},x_{1},x_{2},x_{3} and x4,x5,x6,x7x_{4},x_{5},x_{6},x_{7}.

For reference, the CUDA implementation of the implementation of p←G​cp\leftarrow Gc.

1 __device__ void G_q_fast(
2 const uint32_t enc, uint32_t* out0, uint32_t* out1
3 )
4 {
5 uint32_t even = enc & 0x0F0F0F0F;
6 uint32_t odd = (enc & 0xF0F0F0F0) >> 4;
7
8 uint32_t two_even = even << 1;
9 uint32_t two_odd = odd << 1;
10 uint32_t x0 = even & 0xFF;
11 uint32_t concatenator = (1 << 24) | (1 << 16) | (1 << 8) | 1;
12 uint32_t x0_concat = x0 * concatenator;
13
14 uint32_t temp_even = __vadd4(two_even & 0xFFFFFF00, x0_concat);
15 uint32_t temp_odd = __vadd4(two_odd, x0_concat);
16
17 uint32_t s1 = __dp4a(two_even, (uint32_t)0x01010100, (uint32_t)0);
18 uint32_t s2 = __dp4a(two_odd, (uint32_t)0x01010101, s1);
19
20 *out0 = temp_even;
21 *out1 = temp_odd + s2;
22 }

The computation is done by decomposing GG into a sum of different matrices.

G\displaystyle G =(1000000010200000100020001000002014222222100200001000020010000002)=(1000000010000000100000001000000010000000100000001000000010000000)+(0000000000200000000020000000002000000000000000000000000000000000)\displaystyle=\begin{pmatrix}1&0&0&0&0&0&0&0\\ 1&0&2&0&0&0&0&0\\ 1&0&0&0&2&0&0&0\\ 1&0&0&0&0&0&2&0\\ 1&4&2&2&2&2&2&2\\ 1&0&0&2&0&0&0&0\\ 1&0&0&0&0&2&0&0\\ 1&0&0&0&0&0&0&2\end{pmatrix}=\begin{pmatrix}1&0&0&0&0&0&0&0\\ 1&0&0&0&0&0&0&0\\ 1&0&0&0&0&0&0&0\\ 1&0&0&0&0&0&0&0\\ 1&0&0&0&0&0&0&0\\ 1&0&0&0&0&0&0&0\\ 1&0&0&0&0&0&0&0\\ 1&0&0&0&0&0&0&0\\ \end{pmatrix}+\begin{pmatrix}0&0&0&0&0&0&0&0\\ 0&0&2&0&0&0&0&0\\ 0&0&0&0&2&0&0&0\\ 0&0&0&0&0&0&2&0\\ 0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\end{pmatrix}
+(0000000000000000000000000000000002000000000200000000020000000002)+(0000000000000000000000000000000002222222000000000000000000000000)\displaystyle+\begin{pmatrix}0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\\ 0&2&0&0&0&0&0&0\\ 0&0&0&2&0&0&0&0\\ 0&0&0&0&0&2&0&0\\ 0&0&0&0&0&0&0&2\\ \end{pmatrix}+\begin{pmatrix}0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\\ 0&2&2&2&2&2&2&2\\ 0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\\ 0&0&0&0&0&0&0&0\end{pmatrix}

The next step is to compute pq\frac{p}{q} and the coset of pq\frac{p}{q} we merge both operations into one function.

1 __device__ void decode_nestquant(
2 uint32_t *enc, int *B_local_decode, const int N = 8
3 )
4 {
5 uint32_t inp0, inp1;
6 unsigned int dist[2];
7
8 const uint32_t MASK_FRACPART = 0x1F1F1F1F;
9 const uint32_t MASK_INTPART = 0xE0E0E0E0;
10 const int HALF_CONCAT = 0x10101010;
11 const int TWO = 0x40;
12 const int ONE = 0x20;
13 const int HALF = 0x10;
14
15 G_q_fast(enc[0], &inp0, &inp1);
16
17 int32_t fracPart0 = inp0 & MASK_FRACPART;
18 int32_t fracPart1 = inp1 & MASK_FRACPART;
19
20 int32_t integerPart0 = inp0 & MASK_INTPART;
21 int32_t integerPart1 = inp1 & MASK_INTPART;
22
23 int g0 = (fracPart0 & HALF_CONCAT) << 1;
24 int g1 = (fracPart1 & HALF_CONCAT) << 1;
25
26 int32_t change = ((g0 & ONE) - HALF);
27 int32_t sum1 = get_sum(integerPart0, integerPart1);
28 int32_t two_parity_1 = (sum1 & ONE) >> 4;
29 int32_t fracPart0_1 = __vsub4(fracPart0, (int32_t)(two_parity_1 * change) & 0xFF);
30
31 int32_t sum2 = get_sum(g0, g1) + sum1;
32 int32_t two_parity_2 = (sum2 & ONE) >> 4;
33 int32_t fracPart0_2 = __vadd4(fracPart0, (int32_t)(two_parity_2 * change) & 0xFF);
34
35 dist[0] = __vabsdiffs4(fracPart0, HALF_CONCAT);
36 dist[1] = __vabsdiffs4(fracPart1, HALF_CONCAT);
37 int Delta = get_sum(dist[0], dist[1]);
38
39 int32_t dist00 = (dist[0] & 0xFF);
40 Delta -= (two_parity_1 + two_parity_2) * dist00;
41 Delta += two_parity_1 << 4;
42
43 if (Delta <= TWO) {
44 B_local_decode[0] = __vsub4(fracPart0_1, HALF_CONCAT);
45 B_local_decode[1] = __vsub4(fracPart1, HALF_CONCAT);
46 } else {
47 B_local_decode[0] = __vsub4(fracPart0_2, g0);
48 B_local_decode[1] = __vsub4(fracPart1, g1);
49 }
50 }

Remember that for q=16q=16, we would need to divide pp by 1616. We used a basis for 2​E82E_{8}, and not E8E_{8} so every element is doubled. This is why in our code 12\frac{1}{2} is represented by 16, 11 is represented by 3232, and 22 is represented by 6464. In order to avoid using the round function explicitly (which is not parallelized in hardware) we compute the integer and fractional part of the vector using masks (as can be seen in lines 17-21).

We compute gg in two parts g0g_{0} and g1g_{1} (see lines 23 and 24). This is done by checking that the fractional part is greater or equal to 12\frac{1}{2} (equivalently in our scaled basis, that the fractional part has the fourth LSB set to one). We leverage the fact that

x−(⌊x⌋+0.5)={x}−0.5x-(\lfloor x\rfloor+0.5)=\{x\}-0.5

(lines 44 and 45) and

x−round⁡(x)=⌊x⌋+{x}−(⌊x⌋+g)={x}−gx-\mathrm{round}(x)=\lfloor x\rfloor+\{x\}-(\lfloor x\rfloor+g)=\{x\}-g

(lines 47 and 48). To compute the bitflip, we compute 2​g1−12g_{1}-1 (corresponding to the first element in the vector, see line 26 for reference). We add (subtract) it from fractional part based on the parity check we compute in lines 28 and 32 respectively.

To decode betas in our kernel, we encode the betas as indices to a predefined dictionary (of size 4). Thus, each beta can be represented using 2 bits.

1 int beta1 = (beta_packed >> shift) & 0x3;
2 int beta2 = (beta_packed >> (shift + 2)) & 0x3;
3 int decoded_beta1 = beta_dict[beta1];
4 int decoded_beta2 = beta_dict[beta2];

Encoding and decoding kernels share the same logical structure so that the complex mapping defined for beta is shared (this allows fast contiguous reads of multiple betas at once).

E.1 Runtime comparison of GEMV

Table 4: Runtime comparison of GEMV kernels on an 8192×81928192\times 8192 matrix using an NVIDIA A100 GPU.
Method Time (µs)
Baseline (16 bits) 97
NestQuantM (4.25 bits) 60
QuIP# (2 bits) 38
QuIP# (4 bits) ∼\sim75
int4 uniform 31

The QuIP# computation involves invoking two calls to QuIP# (2 bits), so we extrapolate the running time based on QuIP# (2 bits).

Appendix F Dynamic programming for optimal β\beta

Recall that instead of using one instance of lattice codebook CC, we use a union of codebooks CC scaled to different coefficients. Specifically, our final codebook 𝒞\mathcal{C} is parameterized by coefficients β1≤β2≤…≤βk\beta_{1}\leq\beta_{2}\leq\ldots\leq\beta_{k}, and is equal to:

𝒞=β1​C∪β2​C∪…∪βk​C\mathcal{C}=\beta_{1}C\cup\beta_{2}C\cup\ldots\cup\beta_{k}C

Given a set of 8-vectors to quantize, we can find the set of β\beta that minimizes reconstruction error using a dynamic programming algorithm, which is described in Appendix F.

When quantizing a vector to the ii-th scaled codebook, we could either get a small granular error when the vector is in Vβi​Λ​(0)V_{\beta_{i}\Lambda}(0), or a large overload error otherwise. If we use a codebook with smaller β\beta, we have larger chance of overload error, but the expected magnitude of granular error is smaller due to the volume of Voronoi region being smaller (Figure 7). We can have two strategies for encoding:

(a) Overload error probablity
(b) Granular RMSE for five sampled vectors
Figure 7: Granular and overload error for standard Gaussian vectors, q=16q=16
  1. 1.

    First-β\beta: Use the smallest β\beta, which does not result in an overflow error.

  2. 2.

    Opt-β\beta: Try all the values of β\beta, and choose the one that has the smallest reconstruction MSE.

k 2 4 6 8 10
Opt-β\beta 0.0878 0.0795 0.0708 0.0669 0.0646
First-β\beta 0.0878 0.0798 0.0712 0.0676 0.0656
Table 5: Mean RMSE for reconstructed iid standard Gaussian 8-vectors, q=16q=16, kk betas are uniform on [0,10][0,10].

Even though Opt-β\beta should provide smaller error, the definition of First-β\beta will be useful for us. We can note that the difference between error for Opt-β\beta and First-β\beta is not very siginificant (Table 5). Moreover, First-β\beta can be used to determine the optimal set of βi\beta_{i} to use.

Let we have nn samples v1,v2,…,vnv_{1},v_{2},\ldots,v_{n} from the distribution of vectors we are quantizing, and a large set of betas BB, containing β1<β2<…<βm\beta_{1}<\beta_{2}<\ldots<\beta_{m}, from which we want to take the optimal subset of size kk which minimizes the loss under First-β\beta strategy. For each vector viv_{i} and beta βj\beta_{j} we compute m​s​ei​jmse_{ij} — the MSE if we use βj\beta_{j} to quantize viv_{i} and o​v​e​r​l​o​a​di​joverload_{ij} — whether an overload error occurs in this scenario.

We solve the optimization problem with dynamic programming. Let’s define d​pi​jdp_{ij} be the mimimum sum of MSE we can get if we have to quantize all the vectors which do not yield an overload error for βi\beta_{i}, using βi\beta_{i} and j−1j-1 smaller betas and First-β\beta strategy. If ii is large enough so that no vector has an overflow error on βi\beta_{i}, d​pi​kdp_{ik} has the answer to the problem. To compute the value of d​pi​jdp_{ij}, we can iterate over ss — the index of second largest beta in the set (the largest being βi\beta_{i}). Then, the recalculation works in the following way:

d​pi​j\displaystyle dp_{ij} ←min⁡(d​pi​j,d​ps,j−1+∑p,c​o​n​dpm​s​ep​i)\displaystyle\leftarrow\min\left(dp_{ij},dp_{s,j-1}+\sum_{p,cond_{p}}mse_{pi}\right)
where ​c​o​n​dp\displaystyle\text{where }cond_{p} =o​v​e​r​l​o​a​dp​s∧¬o​v​e​r​l​o​a​dp​i\displaystyle=overload_{ps}\wedge\neg overload_{pi}

By following the transtions in this dynamic programming, we can reconstruct the optimal set of β\beta.

Algorithm 6 Dynamic programming for finding the set of β\beta
1: Input: vectors viv_{i}, beta set BB, m​s​ei​jmse_{ij}, o​v​e​r​l​o​a​di​joverload_{ij}
2: d​pi,j=∞dp_{i,j}=\infty for ii in 0​…​m0\ldots m, jj in 0​…​k0\ldots k
3: f​r​o​mi,j=nullfrom_{i,j}=\text{null} for ii in 0​…​m0\ldots m, jj in 0​…​k0\ldots k
4: d​p0,0=0dp_{0,0}=0
5: for i=1i=1 to mm do
6:   for j=1j=1 to kk do
7:    for s=0s=0 to i−1i-1 do
8:      c​o​n​dp=o​v​e​r​l​o​a​dp​s∧¬o​v​e​r​l​o​a​dp​icond_{p}=overload_{ps}\wedge\neg overload_{pi} for p∈1​…​np\in 1\ldots n
9:      c​o​s​t=∑pc​o​n​dp⋅m​s​ep​icost=\sum_{p}cond_{p}\cdot mse_{pi}
10:      if d​pi​j>d​ps,j−1+c​o​s​tdp_{ij}>dp_{s,j-1}+cost then
11:       d​pi​j←d​ps,j−1+c​o​s​tdp_{ij}\leftarrow dp_{s,j-1}+cost
12:       f​r​o​mi​j←sfrom_{ij}\leftarrow s
13:      end if
14:    end for
15:   end for
16: end for
17: Let p​o​spos is chosen so that βp​o​s\beta_{pos} has no overflow errors
18: result = []
19: for j=kj=k downto 11 do
20:   result.append(p​o​spos)
21:   p​o​s←f​r​o​mp​o​s,jpos\leftarrow from_{pos,j}
22: end for

Appendix G Llama experiment details

We choose the train split of the Wikitext2 [35] dataset as a calibration dataset for computing HH, and evaluate the model on the validation split, computing the perplexity metric. For step 2 in the algorithm (Section 4.6), we select β^=[3.5,4.5,6.0,14.5,25.0]/q\hat{\beta}=[3.5,4.5,6.0,14.5,25.0]/q, because it is the β\beta we get when optimizing them for weight quantization without consideration of LDLQ. The overall universe of betas contains values from 11 to 4040 with spacing ranging from 0.250.25 to 22. For running DP on activations, keys, and values, we run the model on a batch of 66 full-length sequences, which is sufficient for this low-dimensional hyperparameter.

When choosing maximum beta for given distribution, we add a margin of 3.0q\frac{3.0}{q} for weights and 4.0q\frac{4.0}{q} to the maximum beta needed to have 00 overload errors on known data to account for potential overload errors in unknown data. While small number of overload error does not affect perplxity significantly, we still aim to minimize their probability.

When computing perplexity for Wikitext2 with given context length, we average the perplexities for all the positions, which is standard for other works in quantization of LLMs.

Appendix H Ablation studies

We found LDLQ to be useful in improving the quality of quantized model. In table 6, we compare the wikitext2 perplexity of models with and without LDLQ.

Algorithm W W + KV W + KV + A
NestQuant 6.308 6.379 6.633
NestQuant (no LDLQ) 6.528 6.605 6.849
Table 6: Effect of LDLQ on NestQuant (q=14q=14 and k=4k=4) wikitext2 perplexity

While Hadamard matrices from Sylvester construction are commonly used in other works (QuIP#, Quarot), there are multiple ways to construct a fast rotation for the case when dimension is not a power of 22 (such as the down projection in MLP of Llama-3). We tested three possible options for rotation on q=14q=14, k=4k=4, W + KV + A quantization.

Algorithm W + KV + A
Fourier 6.773
S⊗HS\otimes H, SS — orthogonal, HH — Sylvester Hadamard 6.770
H1⊗HH_{1}\otimes H, H1H_{1} — hardcoded Hadamard, HH — Sylvester Hadamard 6.663
Table 7: Effect of rotation on NestQuant (q=14q=14 and k=4k=4) wikitext2 perplexity

H.1 The choice of kk

Figure 8: The perplexity-bitrate scaling of NestQuant with different values of kk, all components of the model (weights, KV cache, activations) are quantized

The value of kk, i.e. the number of scaling coefficients is an important hyperparameter of the algorithm. With an increase of kk, we decrease the quantization error by allowing each vector to be quantized to the lattice point with a proper scaling. However, it increases the bitrate and makes the encoding slower, since we need to try a larger number of scaling coefficients.

We used k=3,4,5,8k=3,4,5,8 to quantize Llama-3-8B across different values of qq, plotting the resulting perplexity against bitrate in Figure 8. We can see that using k=3k=3 leads to a suboptimal performance of the quantization scheme, while the performances of k=4,5,8k=4,5,8 are comparable. In our experiments, we use k=4k=4, because lower having kk results in faster encoding.

Appendix I Results for Llama3.2-1B

Here, we show the results of NestQuant on the newer 1B parameter model LLama3.2-1B. We do experiments in the same setups as for the Llama-3-8B model, computing the wikitext2 perplexity.

q Bits Bits (no zstd) W W + KV W + KV + A
14 3.99 4.06 10.061 10.529 11.197
12 3.76 3.837 10.178 10.862 11.910
10 3.50 3.57 10.377 11.552 14.191
8 3.18 3.25 10.850 13.309 18.710
Table 8: Wikitext2 perplexity of NestQuant quantization of Llama-3.2-1B. The format of the table is the same as in Table 3. The perplexity of non-quantized model is 9.749

Appendix J Results for 3-bit model quantization

We publish the results for 3-bit quantization of weights and activations on small models (Llama-3-8B and Llama-2-7B). We use q=7q=7 and k=4k=4, which results in 2.98 bits per entry.

Bits (W-A-KV) Method Llama-2-7B Llama-3-8B
16-16-16 Floating point 5.47 6.14
4-4-16 NestQuant 5.64 6.56
3-3-16 NestQuant 8.25 6.33